﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GeneralGommunication
{
    public static class ProtocolModbusCommon
    {

        /// <summary>
        /// 把2个byte 转为 uint16
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static UInt16 ToUInt16_Big_Endian(this IEnumerable<byte> value, int startIndex)
        {
            return (UInt16)((value.ElementAt(startIndex) << 8) | value.ElementAt(startIndex + 1));
        }
        /// <summary>
        /// 把2个byte 转为 uint16
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static UInt16 ToUInt16_Little_Endian(this IEnumerable<byte> value, int startIndex)
        {
            return (UInt16)((value.ElementAt(startIndex + 1) << 8) | value.ElementAt(startIndex));
        }
        /// <summary>
        /// 把 uint16  大端排序 转为2个byte
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static byte[] GetBytes_Big_endian(this UInt16 value)
        {
            byte[] b = new byte[2];
            b[0] = (byte)(value >> 8);
            b[1] = (byte)value;
            return b;
        }
        /// <summary>
        /// 把 uint16 小端排序 转为2个byte
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static byte[] GetBytes_Little_Endian(this UInt16 value)
        {
            byte[] b = new byte[2];
            b[0] = (byte)(value);
            b[1] = (byte)(value >> 8);
            return b;
        }

        /// <summary>
        /// CRC16
        /// </summary>
        /// <param name="puchMsg">要进行CRC 校验的消息</param>
        /// <param name="index">消息在pushMsg的位置</param>
        /// <param name="usDataLen">消息字节数</param>
        /// <returns></returns>
        public static UInt16 CRC16(this IEnumerable<byte> puchFrame, int index, int usLen)
        {
            return CRC16_normal(puchFrame, index, usLen);
        }
        public static UInt16 CRC16_normal(this IEnumerable<byte> buf, int index, int len)
        {
            UInt16 CRC_ReturnValue = 0xffff;
            int i = 0, j;

            do
            {

                CRC_ReturnValue ^= buf.ElementAt(index + i);
                i++;
                j = 8;
                do
                {
                    if ((CRC_ReturnValue & 0x01) != 0)
                    {
                        CRC_ReturnValue = (UInt16)((CRC_ReturnValue >> 1) ^ 0xa001);
                    }
                    else
                    {
                        CRC_ReturnValue = (UInt16)(CRC_ReturnValue >> 1);
                    }
                    j--;
                }
                while (j > 0);

                len--;
            }
            while (len > 0);
            return CRC_ReturnValue;
        }

        /// <summary>
        /// CRC16
        /// </summary>
        /// <param name="puchMsg">要进行CRC 校验的消息</param>
        /// <param name="index">消息在pushMsg的位置</param>
        /// <param name="usDataLen">消息字节数</param>
        /// <returns></returns>
        public static UInt16 CRC16_fast(this IEnumerable<byte> puchFrame, int index, int usLen)
        {
            if (usLen <= 0)
                return 0;

            byte uchCRCHi = 0xFF; /* 高CRC字节初始化 */
            byte uchCRCLo = 0xFF; /* 低CRC 字节初始化 */
            int iIndex; /* CRC循环中的索引 */
            int i = index;
            while (true) /* 传输消息缓冲区 */
            {
                iIndex = uchCRCHi ^ puchFrame.ElementAt(i); /* 计算CRC */
                i++;
                uchCRCHi = (byte)(uchCRCLo ^ auchCRCHi[iIndex]);
                uchCRCLo = auchCRCLo[iIndex];

                usLen--;
                if (usLen == 0)
                    break;
            }
            //错！！！ 高低字节错了, 所以要反转
            byte t = uchCRCHi;
            uchCRCHi = uchCRCLo;
            uchCRCLo = t;

            return (UInt16)(uchCRCHi << 8 | uchCRCLo);
        }

        /* CRC 高位字节值表 */
        static byte[] auchCRCHi = new byte[]{
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
            0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
            0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
            0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
            0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
            0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
            0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
            0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
            0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
            0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
            0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
            0x80, 0x41, 0x00, 0xC1, 0x81, 0x40
            };
        /* CRC低位字节值表*/
        static byte[] auchCRCLo = new byte[]{
            0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06,
            0x07, 0xC7, 0x05, 0xC5, 0xC4, 0x04, 0xCC, 0x0C, 0x0D, 0xCD,
            0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09,
            0x08, 0xC8, 0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A,
            0x1E, 0xDE, 0xDF, 0x1F, 0xDD, 0x1D, 0x1C, 0xDC, 0x14, 0xD4,
            0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3,
            0x11, 0xD1, 0xD0, 0x10, 0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3,
            0xF2, 0x32, 0x36, 0xF6, 0xF7, 0x37, 0xF5, 0x35, 0x34, 0xF4,
            0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A,
            0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38, 0x28, 0xE8, 0xE9, 0x29,
            0xEB, 0x2B, 0x2A, 0xEA, 0xEE, 0x2E, 0x2F, 0xEF, 0x2D, 0xED,
            0xEC, 0x2C, 0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26,
            0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 0xA0, 0x60,
            0x61, 0xA1, 0x63, 0xA3, 0xA2, 0x62, 0x66, 0xA6, 0xA7, 0x67,
            0xA5, 0x65, 0x64, 0xA4, 0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F,
            0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68,
            0x78, 0xB8, 0xB9, 0x79, 0xBB, 0x7B, 0x7A, 0xBA, 0xBE, 0x7E,
            0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 0xB4, 0x74, 0x75, 0xB5,
            0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71,
            0x70, 0xB0, 0x50, 0x90, 0x91, 0x51, 0x93, 0x53, 0x52, 0x92,
            0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54, 0x9C, 0x5C,
            0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B,
            0x99, 0x59, 0x58, 0x98, 0x88, 0x48, 0x49, 0x89, 0x4B, 0x8B,
            0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
            0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42,
            0x43, 0x83, 0x41, 0x81, 0x80, 0x40
            };

        static byte[] crc8_table = new byte[]
    {
            0x00,0x31,0x62,0x53,0xc4,0xf5,0xa6,0x97,0xb9,0x88,0xdb,0xea,0x7d,0x4c,0x1f,0x2e,
            0x43,0x72,0x21,0x10,0x87,0xb6,0xe5,0xd4,0xfa,0xcb,0x98,0xa9,0x3e,0x0f,0x5c,0x6d,
            0x86,0xb7,0xe4,0xd5,0x42,0x73,0x20,0x11,0x3f,0x0e,0x5d,0x6c,0xfb,0xca,0x99,0xa8,
            0xc5,0xf4,0xa7,0x96,0x01,0x30,0x63,0x52,0x7c,0x4d,0x1e,0x2f,0xb8,0x89,0xda,0xeb,
            0x3d,0x0c,0x5f,0x6e,0xf9,0xc8,0x9b,0xaa,0x84,0xb5,0xe6,0xd7,0x40,0x71,0x22,0x13,
            0x7e,0x4f,0x1c,0x2d,0xba,0x8b,0xd8,0xe9,0xc7,0xf6,0xa5,0x94,0x03,0x32,0x61,0x50,
            0xbb,0x8a,0xd9,0xe8,0x7f,0x4e,0x1d,0x2c,0x02,0x33,0x60,0x51,0xc6,0xf7,0xa4,0x95,
            0xf8,0xc9,0x9a,0xab,0x3c,0x0d,0x5e,0x6f,0x41,0x70,0x23,0x12,0x85,0xb4,0xe7,0xd6,
            0x7a,0x4b,0x18,0x29,0xbe,0x8f,0xdc,0xed,0xc3,0xf2,0xa1,0x90,0x07,0x36,0x65,0x54,
            0x39,0x08,0x5b,0x6a,0xfd,0xcc,0x9f,0xae,0x80,0xb1,0xe2,0xd3,0x44,0x75,0x26,0x17,
            0xfc,0xcd,0x9e,0xaf,0x38,0x09,0x5a,0x6b,0x45,0x74,0x27,0x16,0x81,0xb0,0xe3,0xd2,
            0xbf,0x8e,0xdd,0xec,0x7b,0x4a,0x19,0x28,0x06,0x37,0x64,0x55,0xc2,0xf3,0xa0,0x91,
            0x47,0x76,0x25,0x14,0x83,0xb2,0xe1,0xd0,0xfe,0xcf,0x9c,0xad,0x3a,0x0b,0x58,0x69,
            0x04,0x35,0x66,0x57,0xc0,0xf1,0xa2,0x93,0xbd,0x8c,0xdf,0xee,0x79,0x48,0x1b,0x2a,
            0xc1,0xf0,0xa3,0x92,0x05,0x34,0x67,0x56,0x78,0x49,0x1a,0x2b,0xbc,0x8d,0xde,0xef,
            0x82,0xb3,0xe0,0xd1,0x46,0x77,0x24,0x15,0x3b,0x0a,0x59,0x68,0xff,0xce,0x9d,0xac
    };

        public static byte CRC8(this IEnumerable<byte> datas, int index, int len)
        {
            byte crc = 0x00;

            while (len > 0)
            {
                byte d = datas.ElementAt(index);
                crc = crc8_table[crc ^ d];
                index++;
                len--;
            }
            return (crc);
        }

        //LRC 计算函数
        //buf 全部求和,取反,+1
        public static byte LRC8(this IEnumerable<byte> datas, int index, int len)
        {
            byte lrc = 0;

            for (int i = 0; i < len; i++)
            {
                lrc += datas.ElementAt(i + index);
            }
            lrc = (byte)(~lrc + 1);
            return lrc;
        }



        #region ASCII 与 HEX 转换
        public static byte? Ascii2Hex(this IEnumerable<byte> ascii, int index)
        {
            string msg = System.Text.Encoding.ASCII.GetString(new byte[] { ascii.ElementAt(index), ascii.ElementAt(index + 1) });
            try
            {
                return (byte)int.Parse(msg, System.Globalization.NumberStyles.HexNumber);
            }
            catch
            {
                return null;
            }



        }
        public static byte[] Ascii2Hex(this IEnumerable<byte> ascii, int index, int len)
        {
            if (len % 2 != 0)
                return null;

            List<byte> hex = new List<byte>();
            for (int i = 0; i < len / 2; i++)
            {
                string msg = System.Text.Encoding.ASCII.GetString(new byte[] { ascii.ElementAt(index + i * 2), ascii.ElementAt(index + i * 2 + 1) });

                if (int.TryParse(msg, System.Globalization.NumberStyles.HexNumber, null, out int d))
                {
                    hex.Add((byte)d);
                }
                else {
                    hex.Add(0);
                }
                
            }
            return hex.ToArray();

        }
        public static byte[] HexToAscii(this IEnumerable<byte> hex)
        {
            return HexToAscii(hex, 0, hex.Count());
        }
        public static byte[] HexToAscii(this IEnumerable<byte> hex, int index, int len)
        {
            string msg = "";
            for (int i = 0; i < len; i++)
            {
                msg += hex.ElementAt(index + i).ToString("X2");
            }
            return System.Text.Encoding.ASCII.GetBytes(msg);
        }
        public static byte[] HexToAscii(this byte hex)
        {
            string msg = "";
            for (int i = 0; i < 1; i++)
            {
                msg += hex.ToString("X2");
            }
            return System.Text.Encoding.ASCII.GetBytes(msg);
        }
        #endregion

        #region 调试输出

 
        /// <summary>
        /// 调试打印出16进制 例如 3A 30 31 31 30 30 30
        /// </summary>
        /// <param name="pack"></param>
        /// <returns></returns>
        public static string ToString_Bytes2Hex(this IEnumerable<byte> pack)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < pack.Count(); i++)
            {
                sb.Append($"{pack.ElementAt(i):X2} ");
            }
            return sb.ToString();
        }
        /// <summary>
        /// 调试打印出ascii 例如 
        /// </summary>
        /// <param name="pack"></param>
        /// <returns></returns>
        public static string ToString_Bytes2Ascii(this IEnumerable<byte> pack)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < pack.Count(); i++)
            {
                byte b = pack.ElementAt(i);
                if (b >= 0x20 && b <= 0x7E)
                {
                    sb.Append((char)b);
                }
                else {
                    sb.Append('▯');
                }
            }
            return sb.ToString();
        }
        public static string ToString_Bytes2HexAndAscii(this IEnumerable<byte> pack)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < pack.Count(); i++)
            {
                byte b = pack.ElementAt(i);
                if (b >= 0x20 && b <= 0x7E)
                {
                    //这个字符可以打印
                    sb.Append($"{b:X2}({(char)b}) ");
                }
                else
                {
                    sb.Append($"{b:X2}( ) ");
                }

            }

            return sb.ToString();
        }
        #endregion

        #region u16 to u32 or u16 to float

        /// <summary>
        /// 大端模式 u16 to u32
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static UInt32 U16ToU32_Big_Endian(this IEnumerable<UInt16> value, int startIndex)
        {
            return (UInt32)((value.ElementAt(startIndex) << 16) | value.ElementAt(startIndex + 1));
        }


        /// <summary>
        /// 大端模式 u16 to float
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static float U16ToF32_Big_Endian(this IEnumerable<UInt16> value, int startIndex)
        {
            byte[] bs = new byte[4];
            bs[3] = (byte)(value.ElementAt(startIndex) >> 8);
            bs[2] = (byte)(value.ElementAt(startIndex));
            bs[1] = (byte)(value.ElementAt(startIndex + 1) >> 8);
            bs[0] = (byte)(value.ElementAt(startIndex + 1));
            float num = BitConverter.ToSingle(bs, 0);
            return num;
        }

        /// <summary>
        /// 大端模式 u32 to u16
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static UInt16[] U32ToU16_Big_Endian(this UInt32 value)
        {
            UInt16[] datas = new UInt16[2];
            datas[0] = (UInt16)(value >> 16);
            datas[1] = (UInt16)(value);
            return datas;
        }

        /// <summary>
        /// 大端模式 float to u16
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static UInt16[] F32ToU16_Big_Endian(this float value)
        {
            var bs = BitConverter.GetBytes(value);

            UInt16[] datas = new UInt16[2];
            datas[0] = (UInt16)(bs[3] << 8 | bs[2]);
            datas[1] = (UInt16)(bs[1] << 8 | bs[0]);
            return datas;
        }



        /// <summary>
        /// 小端模式 u16 to u32
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static UInt32 U16ToU32_Little_Endian(this IEnumerable<UInt16> value, int startIndex)
        {
            return (UInt32)((value.ElementAt(startIndex+1) << 16) | value.ElementAt(startIndex));
        }


        /// <summary>
        /// 小端模式 u16 to float
        /// </summary>
        /// <param name="value"></param>
        /// <param name="startIndex"></param>
        /// <returns></returns>
        public static float U16ToF32_Little_Endian(this IEnumerable<UInt16> value, int startIndex)
        {
            var u32 = U16ToU32_Little_Endian(value, startIndex);
            byte[] bs = BitConverter.GetBytes(u32);
            float num = BitConverter.ToSingle(bs, 0);
            return num;
        }

        /// <summary>
        /// 小端模式 u32 to u16
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static UInt16[] U32ToU16_Little_Endian(this UInt32 value)
        {
            UInt16[] datas = new UInt16[2];
            datas[0] = (UInt16)(value);
            datas[1] = (UInt16)(value >> 16);
            return datas;
        }

        /// <summary>
        /// 小端模式 float to u16
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static UInt16[] F32ToU16_Little_Endian(this float value)
        {
            var bs = BitConverter.GetBytes(value);
            var u32 = BitConverter.ToUInt32(bs, 0);
            UInt16[] datas = U32ToU16_Little_Endian(u32);
            return datas;
        }
        #endregion


    }
}
